@@ -95,14 +95,6 @@ Create_Thunk_List (void)
     return tptr;
 }
 
-#ifdef notused
-static void
-Free_Thunk_List (thunk *list)
-{
-    free(list);
-}
-#endif
-
 static thunk *
 Add_Thunk (thunk *list, thunk t)
 {
@@ -537,7 +529,8 @@ static void
 Read_Mappings (FILE *stream)
 {
     char format_buffer[100];
-    char name[1000], *dformat, *format;
+    char name[1000];
+    const char *dformat, *format;
     int count, c;
     Atom atom;
 
@@ -724,7 +717,7 @@ Format_String (const char *string, int unicode)
 }
 
 static const char *
-Format_Len_String (const char *string, int len)
+Format_Len_String (const char *string, int len, int unicode)
 {
     char *data;
     const char *result;
@@ -736,7 +729,7 @@ Format_Len_String (const char *string, int len)
     memcpy(data, string, len);
     data[len] = '\0';
 
-    result = Format_String(data, 0);
+    result = Format_String(data, unicode);
     free(data);
 
     return result;
@@ -757,10 +750,14 @@ is_utf8_locale (void)
 static const char *
 Format_Icons (const unsigned long *icon, int len)
 {
-    char *result = NULL, *tail = NULL;
+    static char *result = NULL;
+    char *tail = NULL;
     int alloced;
     const unsigned long *end = icon + len / sizeof (unsigned long);
 
+    free(result);
+    result = NULL;
+
     alloced = 0;
 
     while (icon < end)
@@ -901,7 +898,7 @@ Format_Len_Text (const char *string, int len, Atom encoding)
 	*_buf_ptr++ = '\0';
 	return _formatting_buffer;
     } else
-	return Format_Len_String(string, len);
+	return Format_Len_String(string, len, 0);
 }
 
 /*
@@ -985,7 +982,6 @@ Format_Len_Unicode (const char *string, int len)
 {
     char *data;
     const char *result, *error;
-    int len2;
 
     int validity = is_valid_utf8(string, len);
 
@@ -1003,12 +999,12 @@ Format_Len_Unicode (const char *string, int len)
 	    error = "<Invalid UTF-8 string: Unknown error>"; break;
 	}
 
-	result = Format_Len_String(string, len);
-	len2 = strlen(result);
-	data = malloc(len2+1);
+	result = Format_Len_String(string, len, 0);
+	/* result is stored in _formatting_buffer, so make a temporary
+	   copy before we overwrite _formatting_buffer with error */
+	data = strdup(result);
 	if (!data)
 	    Fatal_Error("Out of memory!");
-	memcpy(data, result, len2+1);
 
 	memcpy(_formatting_buffer, error, strlen(error)+1);
 	strcat(_formatting_buffer, data);
@@ -1017,20 +1013,7 @@ Format_Len_Unicode (const char *string, int len)
 	return _formatting_buffer;
     }
 
-    if (!is_utf8_locale())
-	return Format_Len_String(string, len);
-
-    data = malloc(len+1);
-    if (!data)
-	Fatal_Error("Out of memory!");
-
-    memcpy(data, string, len);
-    data[len] = '\0';
-
-    result = Format_String(data, 1);
-    free(data);
-
-    return result;
+    return Format_Len_String(string, len, is_utf8_locale());
 }
 
 /*
@@ -1084,7 +1067,7 @@ Format_Thunk (thunk t, char format_char)
 
     switch (format_char) {
       case 's':
-	return Format_Len_String(t.extra_value, (int)t.value);
+	return Format_Len_String(t.extra_value, (int)t.value, 0);
       case 'u':
 	return Format_Len_Unicode(t.extra_value, (int)t.value);
       case 't':
@@ -1444,9 +1427,15 @@ Get_Window_Property_Data_And_Type (Atom atom,
     unsigned long nitems;
     unsigned long nbytes;
     unsigned long bytes_after;
-    unsigned char *prop;
+    static unsigned char *prop = NULL;
     int status;
-	
+
+    if (prop)
+    {
+	XFree(prop);
+	prop = NULL;
+    }
+
     status = XGetWindowProperty(dpy, target_win, atom, 0, (max_len+3)/4,
 				False, AnyPropertyType, &actual_type,
 				&actual_format, &nitems, &bytes_after,
@@ -1519,8 +1508,8 @@ Show_Prop (const char *format, const char *dformat, const char *prop)
     }
 
     thunks = Break_Down_Property(data, (int)length, type, format, size);
-
     Display_Property(thunks, dformat, format);
+    free(thunks);
 }
 
 static void
@@ -1536,6 +1525,7 @@ Show_All_Props (void)
 	    name = Format_Atom(atoms[i]);
 	    Show_Prop(NULL, NULL, name);
 	}
+	XFree(atoms);
     } else
 	for (i = 0; i < font->n_properties; i++) {
 	    atom = font->properties[i].name;
@@ -1565,11 +1555,11 @@ Handle_Prop_Requests (int argc, char **argv)
 	/* Get overriding formats, if any */
 	if (Is_A_Format(argv[0])) {
 	    format = argv++[0]; argc--;
-	    if (!argc) usage();
+	    if (!argc) usage("format specified without atom");
 	}
 	if (Is_A_DFormat(argv[0])) {
 	    dformat = argv++[0]; argc--;
-	    if (!argc) usage();
+	    if (!argc) usage("dformat specified without atom");
 	}
 
 	/* Get property name */
@@ -1608,7 +1598,7 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
     int size;
     char format_char;
     Atom type = 0;
-    unsigned char *data = NULL;
+    const unsigned char *data = NULL;
     int nelements = 0;
 
     atom = Parse_Atom(propname, False);
@@ -1626,7 +1616,14 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	if (size != 8)
 	    Fatal_Error("can't use format character 's' with any size except 8.");
 	type = XA_STRING;
-	data = (unsigned char *) value;
+	data = (const unsigned char *) value;
+	nelements = strlen(value);
+	break;
+      case 'u':
+	if (size != 8)
+	    Fatal_Error("can't use format character 'u' with any size except 8.");
+	type = XInternAtom(dpy, "UTF8_STRING", False);
+	data = (const unsigned char *) value;
 	nelements = strlen(value);
 	break;
       case 't': {
@@ -1655,11 +1652,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	intvalue = strtoul(tmp, NULL, 0);
 	switch(size) {
 	    case 8:
-	        data8[0] = intvalue; data = (unsigned char *) data8; break;
+	        data8[0] = intvalue; data = (const unsigned char *) data8; break;
 	    case 16:    
-	        data16[0] = intvalue; data = (unsigned char *) data16; break;
+	        data16[0] = intvalue; data = (const unsigned char *) data16; break;
 	    case 32:    
-	        data32[0] = intvalue; data = (unsigned char *) data32; break;
+	        data32[0] = intvalue; data = (const unsigned char *) data32; break;
 	}
 	tmp = strtok(NULL,",");
 	while(tmp != NULL){
@@ -1695,11 +1692,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	intvalue = strtoul(tmp, NULL, 0);
 	switch(size) {
 	    case 8:
-	        data8[0] = intvalue; data = (unsigned char *) data8; break;
+	        data8[0] = intvalue; data = (const unsigned char *) data8; break;
 	    case 16:    
-	        data16[0] = intvalue; data = (unsigned char *) data16; break;
+	        data16[0] = intvalue; data = (const unsigned char *) data16; break;
 	    case 32:    
-	        data32[0] = intvalue; data = (unsigned char *) data32; break;
+	        data32[0] = intvalue; data = (const unsigned char *) data32; break;
 	}
 	tmp = strtok(NULL,",");
 	while(tmp != NULL){
@@ -1740,11 +1737,11 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	type = XA_INTEGER;
 	switch (size) {
 	  case 8:
-	    data8 = boolvalue; data = (unsigned char *) &data8; break;
+	    data8 = boolvalue; data = (const unsigned char *) &data8; break;
 	  case 16:
-	    data16 = boolvalue; data = (unsigned char *) &data16; break;
+	    data16 = boolvalue; data = (const unsigned char *) &data16; break;
 	  case 32: default:
-	    data32 = boolvalue; data = (unsigned char *) &data32; break;
+	    data32 = boolvalue; data = (const unsigned char *) &data32; break;
 	}
 	nelements = 1;
 	break;
@@ -1753,7 +1750,7 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
 	static Atom avalue;
 	avalue = Parse_Atom(value, False);
 	type = XA_ATOM;
-	data = (unsigned char *) &avalue;
+	data = (const unsigned char *) &avalue;
 	nelements = 1;
 	break;
       }
@@ -1774,9 +1771,9 @@ Set_Property (Display *dpy, Window w, const char *propname, const char *value)
  */
 
 void
-usage (void)
+usage (const char *errmsg)
 {
-    static const char help_message[] = 
+    static const char *help_message =
 "where options include:\n"
 "    -grammar                       print out full grammar for command line\n"
 "    -display host:dpy              the X server to contact\n"
@@ -1791,10 +1788,15 @@ usage (void)
 "    -fs filename                   where to look for formats for properties\n"
 "    -frame                         don't ignore window manager frames\n"
 "    -f propname format [dformat]   formats to use for property of given name\n"
-"    -spy                           examine window properties forever\n";
+"    -spy                           examine window properties forever\n"
+"    -version                       print program version\n";
 
 
     fflush (stdout);
+
+    if (errmsg != NULL)
+	fprintf (stderr, "%s: %s\n\n", program_name, errmsg);
+
     fprintf (stderr,
 	     "usage:  %s [-options ...] [[format [dformat]] atom] ...\n\n", 
 	     program_name);
@@ -1832,7 +1834,7 @@ Parse_Format_Mapping (int *argc, char ***argv)
 #define ARGC (*argc)
 #define ARGV (*argv)
 #define OPTION ARGV[0]
-#define NXTOPT if (++ARGV, --ARGC==0) usage()
+#define NXTOPT if (++ARGV, --ARGC==0) usage("insufficent arguments for -format")
     char *type_name, *format, *dformat;
   
     NXTOPT; type_name = OPTION;
@@ -1928,12 +1930,12 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-len")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-len requires an argument");
 	    max_len = atoi(argv[0]);
 	    continue;
 	}
 	if (!strcmp(argv[0], "-formats") || !strcmp(argv[0], "-fs")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-fs requires an argument");
 	    if (!(stream = fopen(argv[0], "r")))
 		Fatal_Error("unable to open file %s for reading.", argv[0]);
 	    Read_Mappings(stream);
@@ -1941,14 +1943,14 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-font")) {
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-font requires an argument");
 	    font = Open_Font(argv[0]);
 	    target_win = -1;
 	    continue;
 	}
 	if (!strcmp(argv[0], "-remove")) {
 	    thunk t = {0};
-	    if (++argv, --argc == 0) usage();
+	    if (++argv, --argc == 0) usage("-remove requires an argument");
 	    t.propname = argv[0];
 	    if (remove_props == NULL) remove_props = Create_Thunk_List();
 	    remove_props = Add_Thunk(remove_props, t);
@@ -1956,7 +1958,7 @@ main (int argc, char **argv)
 	}
 	if (!strcmp(argv[0], "-set")) {
 	    thunk t = {0};
-	    if (argc < 3) usage();
+	    if (argc < 3) usage("insufficent arguments for -set");
 	    t.propname = argv[1];
 	    t.extra_value = argv[2];
 	    argv += 3; argc -= 3;
@@ -1972,11 +1974,20 @@ main (int argc, char **argv)
 	    Parse_Format_Mapping(&argc, &argv);
 	    continue;
 	}
-	usage();
+	if (!strcmp(argv[0], "-version")) {
+	    puts(PACKAGE_STRING);
+	    exit(0);
+	}
+	fprintf (stderr, "%s: unrecognized argument %s\n\n",
+		 program_name, argv[0]);
+	usage(NULL);
     }
 
-    if ((remove_props != NULL || set_props != NULL) && argc > 0)
-	usage();
+    if ((remove_props != NULL || set_props != NULL) && argc > 0) {
+	fprintf (stderr, "%s: unrecognized argument %s\n\n",
+		 program_name, argv[0]);
+	usage(NULL);
+    }
 
     if (target_win == None)
 	target_win = Select_Window(dpy, !frame_only);
